--[ Mod data ]--
VisualHeal = AceLibrary("AceAddon-2.0"):new("AceConsole-2.0", "AceEvent-2.0", "AceDB-2.0");
VisualHeal.Data = 
{
    Name = "VisualHeal",
    VersionMajor = 3,
    VersionMinor = 4,    
};

local BC = AceLibrary("Babble-Class-2.2");
local HC = AceLibrary("HealComm-2.0");

--[ Init/Enable/Disable ]--

function VisualHeal:OnInitialize()
    local _, playerClass = UnitClass('player');
    playerClass = string.lower(playerClass);

    if (playerClass == "druid" or playerClass == "paladin" or playerClass == "priest" or playerClass == "shaman") then
        self.IsHealer = true;
    end

    self.PlayerName = UnitName('player');

    VisualHealBar:SetScale(self.db.profile.HealBarScale);
    VisualHealPlayerBar:SetScale(self.db.profile.PlayerBarScale);
end

function VisualHeal:OnEnable()
    self:RegisterEvent("CHAT_MSG_ADDON");

    -- For Driving the UI
    self:RegisterEvent("UNIT_HEALTH");
    self:RegisterEvent("UNIT_AURA");

    -- Register For HealComm Events
    self:RegisterEvent("HealComm_PlayerStart");
    self:RegisterEvent("HealComm_PlayerStop");
    self:RegisterEvent("HealComm_PlayerDelayed");
    self:RegisterEvent("HealComm_UnitStart");
    self:RegisterEvent("HealComm_UnitStop");
    self:RegisterEvent("HealComm_UnitDelayed");
end

function VisualHeal:OnDisable()
    VisualHeal_VersionInfoFrame:Hide();
    self:HealBarHide();
    self:PlayerBarHide();
	self:UnregisterAllEvents()
end

--[ Settings ]--

VisualHeal:RegisterDB("VisualHealSavedVariables", "VisualHealSavedVariablesPerCharacter");
VisualHeal:RegisterDefaults("profile",
{
    ShowHealBar = true,
    ShowPlayerBar = true,
    PlayerBarScale = 1.0,
    HealBarScale = 1.0
});

function VisualHeal:SetOptionHealBarScale(value)
    local a, b, c, d, e = VisualHealBar:GetPoint();
    local xscale = VisualHealBar:GetScale();
    VisualHealBar:SetScale(value);
    if (not b) then
        VisualHealBar:ClearAllPoints();
        VisualHealBar:SetPoint(a, b, c, d * xscale / value, e * xscale / value);
    end
    VisualHeal.db.profile.HealBarScale = value;
end

function VisualHeal:SetOptionPlayerBarScale(value)
    local a, b, c, d, e = VisualHealPlayerBar:GetPoint();
    local xscale = VisualHealPlayerBar:GetScale();
    VisualHealPlayerBar:SetScale(value);
    if (not b) then
        VisualHealPlayerBar:ClearAllPoints();
        VisualHealPlayerBar:SetPoint(a, b, c, d * xscale / value, e * xscale / value);
    end
    VisualHeal.db.profile.PlayerBarScale = value;
end

VisualHeal:RegisterChatCommand({"/visualheal", "/vh"},
{ 
    type="group",
    args = {
        inquiry = {
            type = "execute",
            name = "Inquiry",
            desc = "Perform a version inquiry in the party/raid",
            func = function()
                if (VisualHeal:InParty() or VisualHeal:InRaid()) then
                    VisualHeal.VersionInfo = {}
                    VisualHeal:VersionInfoFrameUpdate();
                    VisualHeal_VersionInfoFrame:Show();
                    VisualHeal.VersionInfoType = "GROUP";
                    SendAddonMessage("VisualHealInquiryRequest", "", VisualHeal:InBattleground() and "BATTLEGROUND" or "RAID");
                else
                    VisualHeal:Print("You are not in a group");
                end
            end
        },
        guildinquiry = {
            type = "execute",
            name = "Guild Inquiry",
            desc = "Perform a version inquiry in your guild",
            func = function()
                if (IsInGuild()) then
                    GuildRoster();
                    VisualHeal.VersionInfo = {}
                    VisualHeal:VersionInfoFrameUpdate();
                    VisualHeal_VersionInfoFrame:Show();
                    VisualHeal.VersionInfoType = "GUILD";
                    SendAddonMessage("VisualHealInquiryRequest", "", "GUILD");
                else
                    VisualHeal:Print("You are not in a guild");
                end
            end
        },
        show = {
            type = "execute",
            name = "Show",
            desc = "Show the HealBar and PlayerBar",
            func = function ()
                VisualHealPlayerBarTextLeft:SetText("");
                VisualHealPlayerBarTextRight:SetText("");
                VisualHealPlayerBarLeft:SetStatusBarColor(1, 1, 1);
                VisualHealPlayerBarLeft:SetValue(1);
                VisualHealPlayerBarMiddle:SetValue(1);
                VisualHealPlayerBarRight:SetValue(1);
                VisualHealPlayerBarSparkOne:SetPoint("CENTER", "VisualHealPlayerBarLeft", "LEFT", 0, 0)
                VisualHealPlayerBarSparkTwo:SetPoint("CENTER", "VisualHealPlayerBarLeft", "LEFT", 372/2, 0)
                VisualHealPlayerBar:Show();
                if (VisualHeal.IsHealer) then
                    VisualHealBarText:SetText("");
                    VisualHealBarLeft:SetStatusBarColor(1, 1, 1);
                    VisualHealBarLeft:SetValue(1);
                    VisualHealBarMiddle:SetValue(1);
                    VisualHealBarRight:SetValue(1);
                    VisualHealBarSparkOne:SetPoint("CENTER", "VisualHealBarLeft", "LEFT", 0, 0)
                    VisualHealBarSparkTwo:SetPoint("CENTER", "VisualHealBarLeft", "LEFT", 372/2, 0)
                    VisualHealBar:Show();
                end
            end
        },
        healbar = {
            type = "toggle",
            name = "HealBar",
            desc = "Toggles display of the HealBar",
            get = function() return VisualHeal.db.profile.ShowHealBar end,
            set = function() VisualHeal.db.profile.ShowHealBar = not VisualHeal.db.profile.ShowHealBar end,
        },
        playerbar = {
            type = "toggle",
            name = "PlayerBar",
            desc = "Toggles display of the PlayerBar",
            get = function() return VisualHeal.db.profile.ShowPlayerBar end,
            set = function() VisualHeal.db.profile.ShowPlayerBar = not VisualHeal.db.profile.ShowPlayerBar end,
        },
        healbarscale = {
            type = "range",
            name = "Scale of the HealBar",
            desc = "Set the scale of the HealBar",
            min = 0.5,
            max = 2.0,
            isPercent = true,
            get = function() return VisualHeal.db.profile.HealBarScale end,
            set = "SetOptionHealBarScale",
        },
        playerbarscale = {
            type = "range",
            name = "Scale of the PlayerBar",
            desc = "Set the scale of the PlayerBar",
            min = 0.5,
            max = 2.0,
            isPercent = true,
            get = function() return VisualHeal.db.profile.PlayerBarScale end,
            set = "SetOptionPlayerBarScale",
        },
        resetbars = {
            type = "execute",
            name = "Reset Bars",
            desc = "Reset the HealBar and PlayerBar to default positions and scales",
            func = function()
                VisualHealBar:ClearAllPoints();
                VisualHealBar:SetScale(1);
                VisualHeal.db.profile.HealBarScale = 1;
                VisualHealBar:SetPoint("CENTER", CastingBarFrame, "CENTER", 0, VisualHealBar.Offset);
                VisualHealPlayerBar:ClearAllPoints();
                VisualHealPlayerBar:SetScale(1);
                VisualHeal.db.profile.PlayerBarScale = 1;
                VisualHealPlayerBar:SetPoint("CENTER", CastingBarFrame, "CENTER", 0, VisualHealPlayerBar.Offset);
            end
        },
    },
});

function VisualHeal:OnProfileEnable()
    self:SetOptionHealBarScale(VisualHeal.db.profile.HealBarScale);
    self:SetOptionPlayerBarScale(VisualHeal.db.profile.HealBarScale);
end

--[ VersionInfo Functions ]--

function VisualHeal:VersionInfoFrameUpdate()
    local i;
    local name;
    local left, right;
    local count;
    local colorGreen = "|cFF00FF00";
    local colorYellow = "|cFFFFFF00";
    local colorRed = "|cFFFF0000";
    local thisVersion = VisualHeal.Data.VersionMajor .. "." .. VisualHeal.Data.VersionMinor;
    local selfRecent = true;
    local mostRecent = thisVersion;

    for i,v in pairs(self.VersionInfo) do
        if (self:CompareVersions(v, mostRecent) == 1) then
            mostRecent = v;
        end
    end

    if (self:CompareVersions(mostRecent, thisVersion) == 1) then
        selfRecent = false;
    end

    left = "Player Name\n" .. self.PlayerName .. "\n";
    right = "Version\n" .. (selfRecent and colorGreen or colorYellow) .. VisualHeal.Data.VersionMajor .. "." .. VisualHeal.Data.VersionMinor .. "\n";
    count = 1;

    for i,v in pairs(self.VersionInfo) do
        count = count + 1;
        left = left .. i .. "\n";
        right = right .. ((self:CompareVersions(v, mostRecent) == 2) and colorYellow or colorGreen) .. v .. "\n";
    end
    right = right .. colorRed;
    if (self.VersionInfoType == "GROUP") then
        if (self:InRaid()) then
            for i = 1,GetNumRaidMembers() do
                name = UnitName('raid' .. i);
                if (not UnitIsUnit('player', 'raid' .. i)) then
                    if (not self.VersionInfo[name]) then
                        count = count + 1;
                        left = left .. name .. "\n";
                        right = right .. "-" .. "\n";
                    end
                end
            end
        elseif self:InParty() then
            for i = 1,GetNumPartyMembers() do
                name = UnitName('party' .. i);
                if (not self.VersionInfo[name]) then
                    count = count + 1;
                    left = left .. name .. "\n";
                    right = right .. "-" .. "\n";
                end
            end
        end
    elseif (self.VersionInfoType == "GUILD") then
        local guildCount = GetNumGuildMembers(true);
		for i = 1,guildCount do
			local name, rank, rankIndex, level, class, zone, note, officernote, online, status = GetGuildRosterInfo(i);
            if (online) then
                if (not self.VersionInfo[name] and (name ~= self.PlayerName)) then
                    count = count + 1;
                    left = left .. name .. "\n";
                    right = right .. "-" .. "\n";
                end
            end
        end        
    end

    VisualHeal_VersionInfoFrame:SetHeight(12 * count + 52);
    VisualHeal_VersionInfoFrame_TextLeft:SetText(left);
    VisualHeal_VersionInfoFrame_TextRight:SetText(right);
end

--[ PlayerBar Functions ]--

function VisualHeal:PlayerBarShow()
    if (self.db.profile.ShowPlayerBar) then
        self.PlayerModifier = self:GetHealModifier('player');
        self:PlayerBarUpdate();
        VisualHealPlayerBar:Show();
        VisualHealPlayerBar.IsLive = true;
    end
end

function VisualHeal:PlayerBarHide()
    VisualHealPlayerBar.IsLive = false;
    VisualHealPlayerBar:Hide();
end

function VisualHeal:PlayerBarUpdate()
    local hp, hppre, hppost;
    local incomingHeal, nextTimeLeft, nextSize, nextName = HC:GetIncomingHeal(self.PlayerName, GetTime() + 1000);

    if (not incomingHeal) then
        self:PlayerBarHide();
        return;
    end       

    -- Determine health percentages
    hp = UnitHealth('player') / UnitHealthMax('player');
    hppre = hp + (nextSize * self.PlayerModifier) / UnitHealthMax('player');
    hppost = hp + (incomingHeal * self.PlayerModifier) / UnitHealthMax('player');

    -- Sanity check on values
    if (hp > 1.0) then 
        hp = 1.0; 
    end
    if (hppre > 2.0) then 
        hpre = 2.0;
    end
    if (hppost > 3.0) then 
        hppost = 3.0;
    end
    if (hp > hppre) then 
        hppre = hp;
    end
    if (hppre > hppost) then 
        hppost = hppre;
    end   

    -- Update bars
    VisualHealPlayerBarLeft:SetValue(hp);
    VisualHealPlayerBarMiddle:SetValue(hppre);
    VisualHealPlayerBarRight:SetValue(hppost);
    VisualHealPlayerBarSparkOne:SetPoint("CENTER", "VisualHealPlayerBarLeft", "LEFT", 372/2 * hp, 0);
    VisualHealPlayerBarSparkTwo:SetPoint("CENTER", "VisualHealPlayerBarLeft", "LEFT", 372/2 * ((incomingHeal == nextSize) and hp or hppre), 0);

    -- Set colour for health
    local r = hp < 0.5 and 1.0 or 2.0 * (1.0 - hp);
    local g = hp > 0.5 and 0.8 or 1.6 * hp;
    local b = 0.0;
    VisualHealPlayerBarLeft:SetStatusBarColor(r, g, b);

    -- Set text
    VisualHealPlayerBarTextLeft:SetText(nextName);
    VisualHealPlayerBarTextRight:SetText(string.format("%.1fs", nextTimeLeft));

    -- Set colour for remaining incoming heal
    VisualHealPlayerBarRight:SetStatusBarColor(0.4, 0.6, 0.4, 0.5);

    -- Set colour for next heal
    VisualHealPlayerBarMiddle:SetStatusBarColor(0.0, 0.8, 0.0, (1.5 - nextTimeLeft) / 1.5);
end

--[ HealBar Functions ]--

function VisualHeal:HealBarShow()
    if (self.db.profile.ShowHealBar) then
        local _, class = UnitClass(HC.CastInfo.HealingTargetUnitID);
        VisualHealBarText:SetText("|cFF" .. (BC:GetHexColor(class) or "") .. HC.CastInfo.HealingTargetName);
        self.TargetModifier = self:GetHealModifier(HC.CastInfo.HealingTargetUnitID);
        self:HealBarUpdate();
        VisualHealBar:Show();
        VisualHealBar.IsLive = true;
    end
end

function VisualHeal:HealBarHide()
    VisualHealBar.IsLive = false;
    VisualHealBar:Hide();
end

function VisualHeal:HealBarUpdate()
    local hp, hppre, hppost, waste;
    local healingTargetUnitID = HC.CastInfo.HealingTargetUnitID;
    local incomingHeal = HC:GetIncomingHeal(HC.CastInfo.HealingTargetName, HC.CastInfo.Time) or 0;

    -- Determine health percentages of healing target
    if self:UnitHasHealthInfo(healingTargetUnitID) then
        -- Full info available
        hp = UnitHealth(healingTargetUnitID) / UnitHealthMax(healingTargetUnitID);
        hppre = (UnitHealth(healingTargetUnitID) + incomingHeal * self.TargetModifier) / UnitHealthMax(healingTargetUnitID);
        hppost = (UnitHealth(healingTargetUnitID) + (incomingHeal + HC.CastInfo.Size) * self.TargetModifier) / UnitHealthMax(healingTargetUnitID);
    else
        -- Estimate
        hp = UnitHealth(healingTargetUnitID) / 100;
        local factor = hp / self:EstimateUnitHealth(healingTargetUnitID);
        hppre = hp + incomingHeal * self.TargetModifier * factor;
        hppost = hp + (incomingHeal + HC.CastInfo.Size) * self.TargetModifier * factor;
    end

    -- Determine waste (in percent)
    if (hppost > 1.0 and hppost > hppre) then
        waste = (hppost - 1.0) / (hppost - hppre);
    else
        waste = 0.0;
    end
    if (waste > 1.0) then 
        waste = 1.0;
    end

    -- Calculate colour for overheal severity
    local red = waste > 0.1 and 1 or waste * 10;
    local green = waste < 0.1 and 1 or -2.5 * waste + 1.25;
    if (waste < 0.0) then 
        green = 1.0;
        red = 0.0;
    end

    -- Sanity check on values
    if (hp > 1.0) then 
        hp = 1.0; 
    end
    if (hppre > 2.0) then 
        hpre = 2.0;
    end
    if (hppost > 3.0) then 
        hppost = 3.0;
    end
    if (hp > hppre) then 
        hppre = hp;
    end
    if (hppre > hppost) then 
        hppost = hppre;
    end    

    -- Update bars
    VisualHealBarLeft:SetValue(hp);
    VisualHealBarMiddle:SetValue(hppre);
    VisualHealBarRight:SetValue(hppost);
    VisualHealBarSparkOne:SetPoint("CENTER", "VisualHealBarLeft", "LEFT", 372/2 * hp, 0)
    VisualHealBarSparkTwo:SetPoint("CENTER", "VisualHealBarLeft", "LEFT", 372/2 * hppre, 0)

    -- Set colour for health
    VisualHealBarLeft:SetStatusBarColor(hp < 0.5 and 1.0 or 2.0 * (1.0 - hp), hp > 0.5 and 0.8 or 1.6 * hp, 0.0);

    -- Set colour for heal
    VisualHealBarRight:SetStatusBarColor(red, green, 0.0);
end

--[ Event Handlers ]--

function VisualHeal:CHAT_MSG_ADDON(prefix, msg, channel, sender)
    -- filter out own messages
    if (sender == self.PlayerName) then
        return;
    end

    if (prefix == "VisualHealInquiryRequest") then
        SendAddonMessage("VisualHealInquiryConfirm", self.Data.VersionMajor .. ":" .. self.Data.VersionMinor .. ":1", channel);
    elseif (prefix == "VisualHealInquiryConfirm") then
        local _, _, versionMajor, versionMinor, commVersion = string.find(msg, "^(%d+):(%d+):(%d+)");
        if (versionMajor) then
            if (not self.VersionInfo) then
                self.VersionInfo = {};
            end
            self.VersionInfo[sender] = versionMajor .. "." .. versionMinor;
            self:VersionInfoFrameUpdate();
        end           
    end
end

function VisualHeal:UNIT_HEALTH(unit)
    if (VisualHealBar.IsLive and UnitIsUnit(HC.CastInfo.HealingTargetUnitID, unit)) then
        -- Update the heal bar when unit health changes on the healing target
        self:HealBarUpdate();
    end
    if (VisualHealPlayerBar.IsLive and UnitIsUnit('player', unit)) then
        self:PlayerBarUpdate();
    end
end

function VisualHeal:UNIT_AURA(unit)
    if (VisualHealBar.IsLive and UnitIsUnit(HC.CastInfo.HealingTargetUnitID, unit)) then
        local newModifier = self:GetHealModifier(HC.CastInfo.HealingTargetUnitID);
        if (self.TargetModifier ~= newModifer) then
            -- Update the heal bar if the buffs/debuffs of the target changed
            self.TargetModifier = newModifier;
            self:HealBarUpdate();
        end
    end
    if (VisualHealPlayerBar.IsLive and UnitIsUnit('player', unit)) then
        local newModifier = self:GetHealModifier('player');
        if (self.PlayerModifier ~= newModifier) then
            self.PlayerModifier = newModifier;
            self:PlayerBarUpdate();
        end
    end
end

function VisualHeal:HealComm_PlayerStart()
    self:HealBarShow();
end
function VisualHeal:HealComm_PlayerStop()
    self:HealBarHide();
end
function VisualHeal:HealComm_PlayerDelayed()
    if (VisualHealBar.IsLive) then
        self:HealBarUpdate();
    end
end
function VisualHeal:HealComm_UnitStart(unitHealer, unitTarget)
    if (VisualHealBar.IsLive and UnitIsUnit(HC.CastInfo.HealingTargetUnitID, unitTarget)) then
        self:HealBarUpdate(); 
    end
    if (VisualHealPlayerBar.IsLive) then
        self:PlayerBarUpdate();
    elseif (UnitIsUnit(unitTarget, 'player')) then
        self:PlayerBarShow(); 
    end
end
function VisualHeal:HealComm_UnitStop(unit)
    if (VisualHealBar.IsLive) then
        self:HealBarUpdate(); 
    end
    if (VisualHealPlayerBar.IsLive) then
        self:PlayerBarUpdate();
    end
end
function VisualHeal:HealComm_UnitDelayed(unit)
    if (VisualHealBar.IsLive) then
        self:HealBarUpdate();
    end
    if (VisualHealPlayerBar.IsLive) then
        self:PlayerBarUpdate(); 
    end
end
